@@ -1,2 +1,2 @@
-from .cca import simple_cca, incremental_cca
 from .confusion import calculate_direct_confusion
+from .seurat import find_anchor, integrate
